import os
import hashlib
import zipfile
import urllib


def check_md5_hash(file: str, hash: str) -> bool:
    """
    Check the MD5 hash of a file. Can be used on large files, as it reads 4K at
    a time.

    Args:
        file (str): The file to check.
        hash (str): The MD5 hash.

    Returns:
        bool: Whether the hash matches.

    """
    assert os.path.exists(file) and os.path.isfile(file)

    md5_hash = hashlib.md5()
    with open(file, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            md5_hash.update(chunk)

    if md5_hash.hexdigest() == hash:
        return True

    return False


def extract_zip(file: str) -> None:
    """
    Extract a zip file in the current directory.

    Args:
        file (str): The zip file to be extracted.

    """
    assert os.path.exists(file) and zipfile.is_zipfile(file)

    with zipfile.ZipFile(file, 'r') as zip_ref:
        zip_ref.extractall(os.path.dirname(file))


def download_data(url: str, path: str) -> None:
    """
    Download the data specified by url to path.

    Args:
        url (str): URL location of the data.
        path (str): Local path where to save data.

    """
    assert not os.path.exists(path) and os.path.exists(os.path.dirname(path))

    try:
        urllib.request.urlretrieve(url, path)
    except ValueError:
        # not an url or url not found
        pass
