import os

from torch.utils.data import Dataset, DataLoader

from .dataset_utils import *


URL = {
    'train': 'https://huggingface.co/datasets/roneneldan/TinyStories/resolve/main/TinyStoriesV2-GPT4-train.txt?download=true',
    'test': 'https://huggingface.co/datasets/roneneldan/TinyStories/resolve/main/TinyStoriesV2-GPT4-train.txt?download=true',
    'valid': 'https://huggingface.co/datasets/roneneldan/TinyStories/resolve/main/TinyStoriesV2-GPT4-valid.txt?download=true'
}

MD5 = {
    'train': 'f14749c06624ab6311c8e50ca80f4043',
    'test': 'f14749c06624ab6311c8e50ca80f4043',
    'valid': '8caebaf671ccbaaafe769409ef7f63c7',
}

BASEPATH = 'tinystories'

_EXTRACTED_FILES = {
    'train': os.path.join(BASEPATH, 'TinyStoriesV2-GPT4-train.txt'),
    'test': os.path.join(BASEPATH, 'TinyStoriesV2-GPT4-train.txt'),
    'valid': os.path.join(BASEPATH, 'TinyStoriesV2-GPT4-valid.txt'),
}

class TinyStoriesDataset(Dataset):

    def __init__(self, root: str | None = None, split: str = 'train'):
        """
        TinyStories dataset.

        Args:
            root (str | None): Directory where the dataset is saved. Defaults to
                os.path.expanduser('~/.cache/optimus/') if None.
            split (str): Split to be returned. Can be 'train', 'test' or
                'valid'.

        """
        super().__init__()

        assert split in ('train', 'valid', 'test'), \
            "Split must be 'train', 'test' or 'valid'! Aborting..."

        if root is None:
            root = os.path.expanduser('~/.cache/optimus/')

        if not os.path.exists(root):
            os.makedirs(root)

        path = os.path.join(root, _EXTRACTED_FILES[split])

        if not os.path.exists(path):
            print(f"Downloading TinyStories dataset for '{split}' split...")

            if not os.path.exists(os.path.join(root, BASEPATH)):
                os.makedirs(os.path.join(root, BASEPATH))

            download_data(URL[split], path)

            print(f"Download complete. Dataset saved to {path}.")

            if check_md5_hash(path, MD5[split]):
                print("Checksum matches!")
            else:
                # delete downloaded file and force exit on checksum mismatch
                os.remove(path)
                assert 0 != 0, (
                    "Checksum doesn't match! Please check the download URL "
                    "and retry the operation! Aborting..."
                )

        else:
            print(f"Found dataset at '{path}'. Using this for '{split}' split...")

        self.stories = []

        # open the dataset file and read the stories from it
        with open(path, 'r') as file:

            story = []

            for line in file:

                if line == '<|endoftext|>\n':
                    # found the beginning of a story; save the previous one and
                    # begin building a new story
                    self.stories.append(' '.join(story))
                    story = []

                else:
                    # append the line to the story
                    story.append(line)

        train_test_split = int(0.95 * len(self.stories))

        if split == 'train':
            self.stories = self.stories[:train_test_split]
        elif split == 'test':
            self.stories = self.stories[train_test_split:]

    def __len__(self) -> int:
        """
        Return the length of the dataset, which is the total number of
        TinyStories stories contained in it.
        """
        return len(self.stories)

    def __getitem__(self, idx: int) -> str:
        """
        Return the story indexed by idx.

        Args:
            idx (int): The index of the story in the dataset.

        """
        return self.stories[idx]


if __name__=='__main__':
    for split in ['train', 'test', 'valid']:
        dataset = TinyStoriesDataset(split=split)
        print(len(dataset))

        print(f"Sample for '{split}' split:")
        data_loader = DataLoader(dataset, batch_size=1, shuffle=True)
        for data in data_loader:
            print(''.join(data))
            break
