import os
import re

from torch.utils.data import Dataset, DataLoader

from .dataset_utils import *


URL = 'https://s3.amazonaws.com/research.metamind.io/wikitext/wikitext-103-v1.zip'

MD5 = '9ddaacaf6af0710eda8c456decff7832'

BASEPATH = 'wikitext-103'

_EXTRACTED_FILES = {
    'train': os.path.join(BASEPATH, 'wiki.train.tokens'),
    'test': os.path.join(BASEPATH, 'wiki.test.tokens'),
    'valid': os.path.join(BASEPATH, 'wiki.valid.tokens'),
}

class WikiText103Dataset(Dataset):

    def __init__(self, root: str | None = None, split: str = 'train'):
        """
        WikiText103 dataset.

        Args:
            root (str | None): Directory where the dataset is saved. Defaults to
                os.path.expanduser('~/.cache/optimus/') if None.
            split (str): Split to be returned. Can be 'train', 'test' or
                'valid'.

        """
        super().__init__()

        assert split in ('train', 'test', 'valid'), \
            "Split must be 'train', 'test' or 'valid'! Aborting..."

        if root is None:
            root = os.path.expanduser('~/.cache/optimus/')

        if not os.path.exists(root):
            os.makedirs(root)

        path = os.path.join(root, _EXTRACTED_FILES[split])
        zip_path = os.path.join(root, os.path.basename(URL))

        if not os.path.exists(path):
            print(f"Downloading WikiText103 dataset for '{split}' split...")
            download_data(URL, zip_path)
            print(f"Download complete. Dataset saved to {path}.")

            if check_md5_hash(zip_path, MD5):
                print("Checksum matches! Unzipping dataset...")
                extract_zip(zip_path)
                os.remove(zip_path) # remove zip
            else:
                # delete downloaded file and force exit on checksum mismatch
                os.remove(zip_path)
                assert 0 != 0, (
                    "Checksum doesn't match! Please check the download URL "
                    "and retry the operation! Aborting..."
                )
        else:
            print(f"Found dataset at '{path}'. Using this for '{split}' split...")

        self.articles = []

        # we'll read the file line by line, but what we really want to do is
        # have a whole article pieced together; this makes it easier to shuffle
        # the data for training, while still maintaining information intact
        pattern = re.compile(r'^ = [^=]+ = $')

        # open the dataset file and read the articles from it
        with open(path, 'r') as file:

            article = []

            for line in file:

                if pattern.match(line):
                    # found the beginning of an article; save the previous one
                    # and begin building a new article
                    self.articles.append(' '.join(article))
                    article = []

                # append the line to the article
                article.append(line)

        # delete the first empty article
        del self.articles[0]

    def __len__(self) -> int:
        """
        Return the length of the dataset, which is the total number of Wikipedia
        articles contained in it.
        """
        return len(self.articles)

    def __getitem__(self, idx: int) -> str:
        """
        Return the article indexed by idx.

        Args:
            idx (int): The index of the article in the dataset.

        """
        return self.articles[idx]


if __name__=='__main__':
    for split in ['train', 'test', 'valid']:
        dataset = WikiText103Dataset(split=split)
        print(len(dataset))

        print(f"Sample for '{split}' split:")
        data_loader = DataLoader(dataset, batch_size=1, shuffle=True)
        for data in data_loader:
            print(''.join(data))
            break
