import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class Norm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        """
        Root-mean square layer normalization (https://arxiv.org/abs/1910.07467).

        Args:
            See Transformer.

        """
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x):
        std = x.pow(2).mean(dim=-1, keepdim=True).sqrt()
        return x / (std + self.eps)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # compute in float32, not in fp16, since normalization needs to be accurate
        output = self._norm(x.float()).type_as(x)
        return output * self.weight


class Attention(nn.Module):
    def __init__(self, n_heads: int, dim: int, p_drop: float):
        """
        Multi-head attention layer.

        Args:
            See Transformer.

        """
        super().__init__()

        assert dim % n_heads == 0
        self.head_dim = dim // n_heads
        self.heads = n_heads
        self.scale = 1 / math.sqrt(self.head_dim)

        self.wq = nn.Linear(dim, n_heads * self.head_dim, bias=False)
        self.wk = nn.Linear(dim, n_heads * self.head_dim, bias=False)
        self.wv = nn.Linear(dim, n_heads * self.head_dim, bias=False)
        self.wo = nn.Linear(n_heads * self.head_dim, dim, bias=False)

        self.dropout = nn.Dropout(p=p_drop)

    def forward(self,
                x: torch.Tensor,
                mask: torch.Tensor | None = None) -> torch.Tensor:

        bs, seq_len, _ = x.shape # (batch_size, seq_len, dim)

        # linearly project the inputs
        Q = self.wq(x) # (batch_size, seq_len, n_heads * head_dim)
        K = self.wk(x)
        V = self.wv(x)

        # split into n_heads to compute attention
        queries = Q.view(bs, seq_len, self.heads, self.head_dim).permute(0, 2, 1, 3) # (batch_size, n_heads, seq_len, head_dim)
        keys = K.view(bs, seq_len, self.heads, self.head_dim).permute(0, 2, 1, 3)
        values = V.view(bs, seq_len, self.heads, self.head_dim).permute(0, 2, 1, 3)

        # compute attention matmul
        keys = keys.permute(0, 1, 3, 2) # (batch_size, n_heads, head_dim, seq_len)
        scores = queries @ keys # (batch_size, n_heads, seq_len, seq_len)

        # scale
        scores = scores * self.scale

        # mask
        if mask is not None:
            scores = scores + mask

        # softmax (attention probabilities) + dropout
        attn_probs = self.dropout(F.softmax(scores, dim=-1))

        # matmul
        output = attn_probs @ values # (batch_size, n_heads, seq_len, head_dim)

        output = output.permute(0, 2, 1, 3).contiguous()
        output = output.view(bs, seq_len, self.heads * self.head_dim) # (batch_size, seq_len, n_heads * head_dim)

        # final linear
        output = self.wo(output)
        return output


class FeedForward(nn.Module):
    def __init__(self, dim: int, ffn_dim: int):
        """
        Feed forward layer.

        This is a modified version of the original feed forward layer from
        "Attention Is All You Need". The changes are described in "GLU variants
        improve transformer" (https://arxiv.org/abs/2002.05202).

        A summary of the changes proposed:
        - 3 linear layers instead of 2
        - reduce dimension of the 3 matrices to 2/3 * dim to keep computation
          constant
        - change the activation function to SwiGLU (Swish)

        Args:
            See Transformer.

        """
        super().__init__()

        ffn_dim = int(2 * ffn_dim / 3)

        self.fn1 = nn.Linear(dim, ffn_dim, bias=False)
        self.fn2 = nn.Linear(ffn_dim, dim, bias=False)
        self.fn3 = nn.Linear(dim, ffn_dim, bias=False)

    def forward(self, x):
        return self.fn2(F.silu(self.fn1(x)) * self.fn3(x))


class TransformerBlock(nn.Module):
    def __init__(self, n_heads: int, dim: int, p_drop: float):
        """
        A transformer layer, contains a multi-head attention and a feed forward.

        Args:
            See Transformer.

        """
        super().__init__()
        self.attention = Attention(n_heads, dim, p_drop)
        self.ffn = FeedForward(dim, 4 * dim)
        self.attention_norm = Norm(dim)
        self.ffn_norm = Norm(dim)
        self.dropout = nn.Dropout(p_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute a layer of the model (attention + ffn).
        """
        _, seq_len, _ = x.shape # (bs, seq_len, dim)

        # compute mask for masked self-attention
        mask = torch.full((1, seq_len, seq_len), float("-inf"))
        mask = torch.triu(mask, diagonal=1).to(x.device)

        # compute normed attention and normed feed forward
        x = x + self.dropout(self.attention(self.attention_norm(x), mask))
        x = x + self.dropout(self.ffn(self.ffn_norm(x)))
        return x


class OptimusTransformer(nn.Module):
    def __init__(self,
                 vocab_sz: int,
                 dim: int = 512,
                 n_layers: int = 6,
                 n_heads: int = 8,
                 p_drop: float = 0.0,
                 weight_tying: bool = False):
        """
        A transformer implementation. Most of it is derived directly from
        "Attention is All You Need" (see https://arxiv.org/abs/1706.03762),
        with additional changes from recent research.

        Args:
            vocab_sz (int): Vocabulary size.
            dim (int): The dimension of embeddings in the model.
            n_layers (int): The number of layers in the model.
            n_heads (int): The number of attention heads.
            p_drop (float): Dropout probability. Dropout is applied to input
                embeddings, the outputs of attention layers, and the outputs of
                feed forward layers.
            weight_tying (bool): Whether to use weight-tying. Intuitively,
                weight-tying could be useful, because the set of mappings from
                the input tokens to the embeddings should be the same as that
                from the embeddings to the output tokens (because the vocabulary
                is the same for both encoding and decoding).

        """
        super().__init__()

        # add as buffers so they get saved along with the model with torch.save()
        self.register_buffer('vocab_sz', torch.tensor(vocab_sz))
        self.register_buffer('n_layers', torch.tensor(n_layers))
        self.register_buffer('n_heads', torch.tensor(n_heads))
        self.register_buffer('dim', torch.tensor(dim))
        self.register_buffer('p_drop', torch.tensor(p_drop))
        self.register_buffer('weight_tying', torch.tensor(weight_tying))

        self.embeddings = nn.Embedding(vocab_sz, dim)
        self.positional_encodings = self._compute_freqs(dim)
        self.input_dropout = nn.Dropout(p=p_drop)
        self.layers = nn.ModuleList([TransformerBlock(n_heads, dim, p_drop) for _ in range(n_layers)])
        self.output = nn.Linear(dim, vocab_sz, bias=False)
        self.output_norm = Norm(dim)

        if weight_tying:
            self.output.weight = self.embeddings.weight

    def _compute_freqs(self, dim: int,
                       max_seq_len: int = 4096,
                       theta: float = 10000.0) -> torch.Tensor:
        """
        Precompute a frequency matrix to apply to input embeddings (positional
        encoding).

        Args:
            dim (int): Dimension of embeddings in the model.
            max_seq_len (int): The maximum context length to precompute
                frequencies for. Defaults to 4096. Needs to be increased for
                bigger context lengths.
            theta (float): Frequency.

        Returns:
            torch.Tensor: Positional encodings tensor.

        """
        pos = torch.arange(0, max_seq_len).unsqueeze(1)
        i = torch.arange(0, dim, 2)
        div_term = torch.exp(i * (-math.log(theta) / dim))

        enc = torch.zeros(max_seq_len, dim)
        enc[:, 0::2] = torch.sin(pos * div_term)
        enc[:, 1::2] = torch.cos(pos * div_term)
        enc = enc.unsqueeze(0)

        return enc

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the model to the input tokens.

        Args:
            x (torch.Tensor): Input token indices. Expected in the shape
                (batch_size, seq_len).

        Returns:
            torch.Tensor: Output logits of the model.

        """
        # input embeddings and positional encoding
        x = self.embeddings(x) # (batch_size, seq_len, dim)
        encs = self.positional_encodings[:,: x.shape[1]].to(x.device)
        x = self.input_dropout(x + encs)

        # go through all model layers
        for layer in self.layers:
            x = layer(x) # (batch_size, seq_len, dim)

        # final norm + linear layer
        x = self.output_norm(x)
        output = self.output(x) # (batch_size, seq_len, vocab_sz)

        return output
